'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var _uniq = _interopDefault(require('lodash/uniq'));
var dependencyGraph = require('dependency-graph');
var Emittery = _interopDefault(require('emittery'));
var graphql = require('graphql');
var ky = require('ky-universal');
var ky__default = _interopDefault(ky);
var pRetry = _interopDefault(require('p-retry'));
var hash = _interopDefault(require('@sindresorhus/fnv1a'));
var _cloneDeep = _interopDefault(require('lodash/cloneDeep'));
var _isPlainObject = _interopDefault(require('lodash/isPlainObject'));
var React = require('react');
var React__default = _interopDefault(React);
var _merge = _interopDefault(require('lodash/merge'));
var AbortController = _interopDefault(require('abort-controller'));
var _mergeWith = _interopDefault(require('lodash/mergeWith'));

class GraphQLError extends Error {
  constructor(errors) {
    super(errors[0].message || 'Request failed');
    this.name = 'GraphQLError';
    this._errors = errors;
  }

  toArray() {
    return this._errors;
  }

  *[Symbol.iterator]() {
    for (const error of this._errors) {
      yield error;
    }
  }

}

class NetworkError extends Error {
  constructor(originalError) {
    super(originalError.message);
    this.name = 'NetworkError';
    this.originalError = originalError;
  }

}

const hasTypename = node => {
  const typenameField = node.selections.find(selection => {
    return selection.name.value === '__typename';
  });
  return Boolean(typenameField);
};

const addTypename = node => {
  node.selections.push({
    kind: graphql.Kind.FIELD,
    name: {
      kind: graphql.Kind.NAME,
      value: '__typename'
    },
    arguments: [],
    directives: []
  });
};

var addTypenames = (query => {
  const queryClone = _cloneDeep(query);

  graphql.visit(queryClone, {
    SelectionSet(node) {
      if (!hasTypename(node)) {
        addTypename(node);
      }
    }

  });
  return queryClone;
});

const scanTypenames = data => {
  if (!data) {
    return [];
  }

  const typenames = [];

  for (const [key, value] of Object.entries(data)) {
    if (key === '__typename') {
      typenames.push(value);
    }

    if (Array.isArray(value)) {
      for (const item of value) {
        typenames.push(...scanTypenames(item));
      }
    }

    if (_isPlainObject(value)) {
      typenames.push(...scanTypenames(value));
    }
  }

  return typenames;
};

var getTypenames = (data => {
  const rootTypename = data.__typename;
  return _uniq(scanTypenames(data)).filter(typename => typename !== rootTypename);
});

class Draqula {
  constructor(url, options = {}) {
    this.url = url;
    this.options = options;
    this.events = new Emittery();
    this.graph = new dependencyGraph.DepGraph();
    this.queryCache = new WeakMap();
    this.queryIds = new WeakMap();
    this.dataCache = new Map();
  }

  createQuery(query) {
    if (this.queryCache.has(query)) {
      return this.queryCache.get(query);
    }

    const queryString = graphql.print(addTypenames(query));
    this.queryCache.set(query, queryString);
    const queryId = String(hash(queryString));
    this.queryIds.set(query, queryId);
    this.dataCache.set(queryId, new Map());
    return queryString;
  }

  getQueryId(query) {
    return this.queryIds.get(query);
  }

  getCacheKey(variables) {
    return JSON.stringify(variables);
  }

  setDataCache(query, variables, data) {
    if (this.options.cache === false) {
      return;
    }

    const queryId = this.getQueryId(query);
    const cache = this.dataCache.get(queryId);

    if (cache) {
      const cacheKey = this.getCacheKey(variables);
      cache.set(cacheKey, data);
    }
  }

  getDataCache(query, variables) {
    if (this.options.cache === false) {
      return;
    }

    const queryId = this.getQueryId(query);
    const cache = this.dataCache.get(queryId);

    if (!cache) {
      return;
    }

    const cacheKey = this.getCacheKey(variables);
    return cache.get(cacheKey);
  }

  async rawQuery(query, variables, options) {
    const execute = async () => {
      try {
        const response = await ky__default.post(this.url, {
          json: {
            query: this.createQuery(query),
            variables
          },
          signal: options.signal,
          hooks: this.options.hooks,
          timeout: options.timeout || this.options.timeout
        }).json();

        if (Array.isArray(response.errors)) {
          const errors = response.errors.map(error => {
            const error2 = new Error(error.message);

            for (const key in error) {
              if ({}.hasOwnProperty.call(error, key)) {
                error2[key] = error[key];
              }
            }

            return error2;
          });
          throw new GraphQLError(errors);
        }

        return response.data;
      } catch (error) {
        const isClientError = error.response && String(error.response.status).startsWith('4');
        const isAborted = error.name === 'AbortError';

        if (isClientError || isAborted) {
          throw new pRetry.AbortError(error);
        }

        throw error;
      }
    };

    let retries = 2;

    if (typeof this.options.retry === 'number') {
      retries = this.options.retry;
    }

    if (typeof this.options.retry === 'object' && typeof this.options.retry.retries === 'number') {
      retries = this.options.retry.retries;
    }

    if (options.retry === false || this.options.retry === false) {
      retries = 0;
    }

    if (retries === 0) {
      return execute();
    }

    try {
      return await pRetry(execute, {
        retries
      });
    } catch (error) {
      if (error.name === ky.HTTPError.name || error.name === ky.TimeoutError.name) {
        throw new NetworkError(error);
      }

      throw error;
    }
  }

  async query(query, variables, options) {
    const data = await this.rawQuery(query, variables, options);

    if (options.cache !== false && this.options.cache !== false) {
      const queryId = this.getQueryId(query);
      this.graph.addNode(`query:${queryId}`);
      const typenames = getTypenames(data);

      for (const typename of typenames) {
        this.graph.addNode(`type:${typename}`);
        this.graph.addDependency(`query:${queryId}`, `type:${typename}`);
      }
    }

    return data;
  }

  async preload(query, variables = {}) {
    const data = await this.query(query, variables, {});

    if (this.options.cache !== false) {
      this.setDataCache(query, variables, data);
    }
  }

  watchQuery(query, callback) {
    const queryId = this.getQueryId(query);
    this.events.on(`refetch:${queryId}`, callback);
    return () => this.events.off(`refetch:${queryId}`, callback);
  }

  refetchQueries(queryIds) {
    return _uniq(queryIds).map(queryId => {
      this.dataCache.delete(queryId);
      return this.events.emit(`refetch:${queryId}`);
    });
  }

  async mutate(query, variables, options) {
    const data = await this.rawQuery(query, variables, {
      retry: false
    });

    if (options.refetchQueries === false) {
      return data;
    }

    const typenames = getTypenames(data);
    const refetchQueries = [];

    for (const typename of typenames) {
      const nodeName = `type:${typename}`;

      if (this.graph.hasNode(nodeName)) {
        const dependantQueries = this.graph.dependantsOf(nodeName).map(queryId => queryId.replace('query:', ''));
        refetchQueries.push(...dependantQueries);
      }
    }

    if (Array.isArray(options.refetchQueries)) {
      const queryIds = options.refetchQueries.map(query => this.getQueryId(query));
      refetchQueries.push(...queryIds);
    }

    const promises = this.refetchQueries(refetchQueries);

    if (options.waitForRefetchQueries) {
      await Promise.all(promises);
    }

    return data;
  }

  clearCache() {
    this.graph = new dependencyGraph.DepGraph();
    this.queryCache = new WeakMap();
    this.queryIds = new WeakMap();
    this.dataCache.clear();
  }

  destroy() {
    this.events.clearListeners();
    this.clearCache();
  }

}

var DraqulaContext = React.createContext({
  client: undefined
});

const DraqulaProvider = ({
  client,
  children
}) => {
  React.useEffect(() => {
    return () => {
      if (client) {
        client.destroy();
      }
    };
  }, [client]);
  return React__default.createElement(DraqulaContext.Provider, {
    value: {
      client
    }
  }, children);
};

var useDraqulaClient = (() => {
  const context = React.useContext(DraqulaContext);

  if (!context || !context.client) {
    throw new Error("Draqula client couldn't be found. Did you forget to pass client to your React components via DraqulaProvider?");
  }

  return context.client;
});

const defaultMutationOptions = {
  refetchQueries: [],
  waitForRefetchQueries: false
};

const reducer = (state, action) => {
  if (action.type === 'mutate') {
    return {
      data: undefined,
      error: undefined,
      isLoading: true
    };
  }

  if (action.type === 'success') {
    return {
      data: action.data,
      error: undefined,
      isLoading: false
    };
  }

  if (action.type === 'error') {
    return {
      data: undefined,
      error: action.error,
      isLoading: false
    };
  }

  return state;
};

const initialState = {
  data: undefined,
  error: undefined,
  isLoading: false
};
var useMutation = ((query, options = defaultMutationOptions) => {
  const client = useDraqulaClient();
  const [{
    data,
    error,
    isLoading
  }, originalDispatch] = React.useReducer(reducer, initialState);
  const isUnmountedRef = React.useRef(false);
  React.useEffect(() => {
    return () => {
      isUnmountedRef.current = false;
    };
  }, []);

  const dispatch = action => {
    if (isUnmountedRef.current === false) {
      originalDispatch(action);
    }
  };

  const mutate = React.useCallback(async (variables = {}) => {
    dispatch({
      type: 'mutate'
    });

    try {
      const data = await client.mutate(query, variables, options);
      dispatch({
        type: 'success',
        data
      });
      return data;
    } catch (error) {
      dispatch({
        type: 'error',
        error
      });
      throw error;
    }
  }, [client, query, JSON.stringify(options)]);
  return {
    mutate,
    data,
    error,
    isLoading
  };
});

var useDataCache = ((query, variables) => {
  const client = useDraqulaClient();
  const cache = React.useMemo(() => {
    return client.getDataCache(query, variables);
  }, [client, query, JSON.stringify(variables)]);
  const setCache = React.useCallback(data => {
    client.setDataCache(query, variables, data);
  }, [client]);
  return [cache, setCache];
});

var usePageFocus = ((callback, options) => {
  React.useEffect(() => {
    if (typeof document === 'undefined' || !options.isEnabled) {
      return;
    }

    const handleVisibilityChange = () => {
      if (!document.hidden) {
        callback();
      }
    };

    document.addEventListener('visibilitychange', handleVisibilityChange);
    return () => {
      document.removeEventListener('visibilitychange', handleVisibilityChange);
    };
  }, [options.isEnabled, callback]);
});

var defaultMerge = ((prevData, nextData) => {
  return _cloneDeep(_mergeWith(prevData, nextData, (objValue, srcValue) => {
    if (!Array.isArray(objValue)) {
      return;
    }

    return objValue.concat(srcValue);
  }));
});

const defaultFetchMoreOptions = {
  merge: defaultMerge
};

const reducer$1 = (state, action) => {
  if (action.type === 'fetch') {
    return { ...state,
      error: undefined,
      isLoading: true
    };
  }

  if (action.type === 'success') {
    return { ...state,
      data: action.data,
      error: undefined,
      isLoading: false
    };
  }

  if (action.type === 'error') {
    return { ...state,
      data: undefined,
      error: action.error,
      isLoading: false
    };
  }

  if (action.type === 'reset') {
    return {
      data: action.data,
      error: undefined,
      isLoading: !action.data,
      isFetchingMore: false
    };
  }

  if (action.type === 'fetch-more') {
    return { ...state,
      isFetchingMore: true
    };
  }

  if (action.type === 'fetch-more-success') {
    return { ...state,
      data: action.data
    };
  }

  if (action.type === 'fetch-more-done') {
    return { ...state,
      isFetchingMore: false
    };
  }

  if (action.type === 'set-data') {
    return { ...state,
      data: action.data
    };
  }

  return state;
};

var useQuery = ((query, variables = {}, options = {}) => {
  const client = useDraqulaClient();
  const [cache, setCache] = useDataCache(query, variables);
  const [{
    data,
    error,
    isLoading,
    isFetchingMore
  }, dispatch] = React.useReducer(reducer$1, {
    data: cache,
    error: undefined,
    isLoading: cache === undefined,
    isFetchingMore: false
  });
  const [customData, setCustomData] = React.useState(cache);
  React.useEffect(() => {
    if (options.cache !== false) {
      setCustomData(data);
    }
  }, [data, options.cache]);
  React.useEffect(() => {
    if (options.cache !== false) {
      dispatch({
        type: 'set-data',
        data: customData
      });
      setCache(customData);
    }
  }, [customData, options.cache]);
  React.useEffect(() => {
    const abortController = new AbortController();

    const fetch = async () => {
      if (cache === undefined) {
        dispatch({
          type: 'fetch'
        });
      }

      try {
        const data = await client.query(query, variables, { ...options,
          signal: abortController.signal
        });
        dispatch({
          type: 'success',
          data
        });
      } catch (error) {
        if (error.name === 'AbortError') {
          return;
        }

        dispatch({
          type: 'error',
          error
        });
      }
    };

    dispatch({
      type: 'reset',
      data: cache
    });
    fetch();
    return () => {
      abortController.abort();
    };
  }, [client, query, JSON.stringify(variables), JSON.stringify(options), cache]);
  const refetchAbortControllerRef = React.useRef();
  const refetch = React.useCallback(async () => {
    if (refetchAbortControllerRef.current) {
      refetchAbortControllerRef.current.abort();
    }

    refetchAbortControllerRef.current = new AbortController();

    try {
      const data = await client.query(query, variables, { ...options,
        signal: refetchAbortControllerRef.current.signal
      });
      dispatch({
        type: 'success',
        data
      });
    } catch (error) {
      if (error.name === 'AbortError') {
        return;
      }

      throw error;
    }
  }, [client, query, JSON.stringify(variables), JSON.stringify(options)]);
  usePageFocus(refetch, {
    isEnabled: typeof options.refetchOnFocus === 'boolean' ? options.refetchOnFocus : true
  });
  const fetchMore = React.useCallback(async (overrideVariables, fetchMoreOptions = defaultFetchMoreOptions) => {
    dispatch({
      type: 'fetch-more'
    });

    try {
      const nextData = await client.query(query, _merge({}, variables, overrideVariables), options);

      if (data === undefined) {
        dispatch({
          type: 'fetch-more-success',
          data: nextData
        });
      }

      if (data !== undefined && nextData !== undefined) {
        dispatch({
          type: 'fetch-more-success',
          data: fetchMoreOptions.merge(data, nextData)
        });
      }
    } finally {
      dispatch({
        type: 'fetch-more-done'
      });
    }
  }, [client, query, JSON.stringify(variables), JSON.stringify(options), data]);
  React.useEffect(() => {
    const unsubscribe = client.watchQuery(query, refetch);
    return unsubscribe;
  }, [client, query, refetch]);
  return {
    data,
    setData: setCustomData,
    error,
    isLoading,
    fetchMore,
    isFetchingMore,
    refetch
  };
});

exports.Draqula = Draqula;
exports.DraqulaContext = DraqulaContext;
exports.DraqulaProvider = DraqulaProvider;
exports.GraphQLError = GraphQLError;
exports.NetworkError = NetworkError;
exports.useDraqulaClient = useDraqulaClient;
exports.useMutation = useMutation;
exports.useQuery = useQuery;
